# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import asyncio
import aiomysql
import sqlalchemy

from abc import ABCMeta, abstractmethod

from aiomysql.sa import SAConnection, Engine
from aiomysql.sa.engine import _dialect as dialect

from pymysql.err import Warning, DataError, IntegrityError, ProgrammingError

from sqlalchemy.sql.selectable import Select
from sqlalchemy.sql.dml import Insert, Update, Delete

from .base import Utils, AsyncCirculator

from ..error import MySQLReadOnlyError, MySQLClientDestroyed
from ..interface import AsyncContextManager


MYSQL_ERROR_RETRY_COUNT = 0x10
MYSQL_POLL_WATER_LEVEL_WARNING_LINE = 0x10


class MySQLPool:
    """MySQL连接管理
    """

    class _Connection(SAConnection):

        def __init__(self, connection, engine, compiled_cache=None):

            super().__init__(connection, engine, compiled_cache)

            if not hasattr(connection, r'build_time'):
                setattr(connection, r'build_time', Utils.loop_time())

        @property
        def build_time(self):

            return getattr(self._connection, r'build_time', 0)

        async def destroy(self):

            if self._connection is None:
                return

            if self._transaction is not None:
                await self._transaction.rollback()
                self._transaction = None

            self._connection.close()

            self._engine.release(self)
            self._connection = None
            self._engine = None

    def __init__(
            self, host, port, db, user, password,
            *, name=None, minsize=8, maxsize=32, echo=False, pool_recycle=21600,
            charset=r'utf8', autocommit=True, cursorclass=aiomysql.DictCursor,
            readonly=False, conn_life=43200,
            **settings
    ):

        self._name = name if name is not None else (Utils.uuid1()[:8] + (r'_ro' if readonly else r'_rw'))
        self._pool: aiomysql.pool.Pool = None
        self._engine: Engine = None
        self._readonly = readonly
        self._conn_life = conn_life

        self._settings = settings

        self._settings[r'host'] = host
        self._settings[r'port'] = port
        self._settings[r'db'] = db

        self._settings[r'user'] = user
        self._settings[r'password'] = password

        self._settings[r'minsize'] = minsize
        self._settings[r'maxsize'] = maxsize

        self._settings[r'echo'] = echo
        self._settings[r'pool_recycle'] = pool_recycle
        self._settings[r'charset'] = charset
        self._settings[r'autocommit'] = autocommit
        self._settings[r'cursorclass'] = cursorclass

    @property
    def name(self):

        return self._name

    @property
    def readonly(self):

        return self._readonly

    @property
    def conn_life(self):

        return self._conn_life

    def __await__(self):

        self._pool = yield from aiomysql.create_pool(**self._settings).__await__()
        self._engine = Engine(dialect, self._pool)

        Utils.log.info(
            f"MySQL [{self._settings[r'host']}:{self._settings[r'port']}] {self._settings[r'db']}"
            f" ({self._name}) initialized: {self._pool.size}/{self._pool.maxsize}"
        )

        return self

    async def close(self):

        if self._pool is not None:

            self._pool.close()
            await self._pool.wait_closed()

            self._pool = None

    def _echo_pool_info(self):

        if self.health:
            Utils.log.debug(
                f'MySQL connection pool info ({self._name}): '
                f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
            )
        else:
            Utils.log.warning(
                f'MySQL connection pool not enough ({self._name}): '
                f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
            )

    @property
    def health(self):

        global MYSQL_POLL_WATER_LEVEL_WARNING_LINE

        return (self._pool.maxsize - self._pool.size + self._pool.freesize) > MYSQL_POLL_WATER_LEVEL_WARNING_LINE

    async def reset(self):

        if self._pool is not None:

            await self._pool.clear()

            Utils.log.info(
                f'MySQL connection pool reset ({self._name}): {self._pool.size}/{self._pool.maxsize}'
            )

    async def get_sa_conn(self):

        self._echo_pool_info()

        conn = await self._pool.acquire()

        return self._Connection(conn, self._engine)

    def get_client(self):

        return DBClient(self)

    def get_transaction(self):

        if self._readonly:
            raise MySQLReadOnlyError()

        return DBTransaction(self)


class MySQLDelegate:
    """MySQL功能组件
    """

    def __init__(self):

        self._mysql_rw_pool = None
        self._mysql_ro_pool = None

    @property
    def mysql_rw_pool(self):

        return self._mysql_rw_pool

    @property
    def mysql_ro_pool(self):

        return self._mysql_ro_pool

    async def init_mysql_rw(self, *args, **kwargs):

        self._mysql_rw_pool = await MySQLPool(*args, **kwargs)

    async def init_mysql_ro(self, *args, **kwargs):

        self._mysql_ro_pool = await MySQLPool(*args, **kwargs)

    async def close_mysql(self):

        if self._mysql_rw_pool is not None:
            await self._mysql_rw_pool.close()

        if self._mysql_ro_pool is not None:
            await self._mysql_ro_pool.close()

    async def mysql_health(self):

        result = await self._check_health(self._mysql_rw_pool)
        result &= await self._check_health(self._mysql_ro_pool)

        return result

    async def _check_health(self, pool):

        if pool is None:
            return True

        result = False

        async with pool.get_client() as client:
            await client.safe_execute(r'select version();')
            result = True

        return result

    async def reset_mysql_pool(self):

        if self._mysql_rw_pool:
            await self._mysql_rw_pool.reset()

        if self._mysql_ro_pool:
            await self._mysql_ro_pool.reset()

    def get_db_client(self, readonly=False):

        if readonly:
            if self._mysql_ro_pool:
                return self._mysql_ro_pool.get_client()
            else:
                return self._mysql_rw_pool.get_client()
        else:
            return self._mysql_rw_pool.get_client()

    def get_db_transaction(self):

        return self._mysql_rw_pool.get_transaction()


class _ClientBase(metaclass=ABCMeta):
    """MySQL客户端基类
    """

    @staticmethod
    def safestr(val, charset=r'utf-8'):

        cls = type(val)

        if cls is str:
            val = aiomysql.escape_string(val)
        elif cls is dict:
            val = aiomysql.escape_dict(val, charset)
        else:
            val = str(val)

        return val

    def __init__(self, pool):

        self._pool = pool

        self._conn = None
        self._trx = None

        self._readonly = pool.readonly

        self._lock = asyncio.Lock()

    @property
    def readonly(self):

        return self._readonly

    @property
    def insert_id(self):

        if self._conn:
            return self._conn.connection.insert_id()
        else:
            return None

    async def _get_conn(self):

        if self._pool is None:
            raise MySQLClientDestroyed()

        if self._conn is None:
            self._conn = await self._pool.get_sa_conn()

        return self._conn

    async def _close_conn(self, discard=False):

        if self._conn is not None:

            _conn, self._conn = self._conn, None

            if discard:
                await _conn.destroy()
            elif (Utils.loop_time() - _conn.build_time) > self._pool.conn_life:
                await _conn.destroy()
            else:
                await _conn.close()

    @abstractmethod
    async def _execute(self, clause, *multiparams, **params):

        raise NotImplementedError()

    async def execute(self, clause):

        result = None

        async with self._lock:
            result = await self._execute(clause)

        return result

    async def safe_execute(self, clause):

        async with self._lock:

            proxy = await self._execute(clause)
            await proxy.close()

            await self._close_conn()

    async def select(self, query, *multiparams, **params):

        result = []

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        async with self._lock:

            proxy = await self._execute(query, *multiparams, **params)

            if proxy is not None:

                records = await proxy.cursor.fetchall()

                if records:
                    result.extend(records)

                if not proxy.closed:
                    await proxy.close()

            if not self._trx:
                await self._close_conn()

        return result

    async def find(self, query, *multiparams, **params):

        result = None

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        async with self._lock:

            proxy = await self._execute(query.limit(1), *multiparams, **params)

            if proxy is not None:

                record = await proxy.cursor.fetchone()

                if record:
                    result = record

                if not proxy.closed:
                    await proxy.close()

            if not self._trx:
                await self._close_conn()

        return result

    async def count(self, table_col, select_where=None):

        _select = sqlalchemy.select([sqlalchemy.func.count(table_col).label(r'tbl_row_count')])

        if select_where is not None:
            _select = _select.where(select_where)

        result = await self.find(_select)

        return result[r'tbl_row_count']

    async def insert(self, query, *multiparams, **params):

        result = 0

        if self._readonly:
            raise MySQLReadOnlyError()

        if not isinstance(query, Insert):
            raise TypeError(r'Not sqlalchemy.sql.dml.Insert object')

        async with self._lock:

            proxy = await self._execute(query, *multiparams, **params)

            if proxy is not None:

                result = self.insert_id

                if not proxy.closed:
                    await proxy.close()

            if not self._trx:
                await self._close_conn()

        return result

    async def update(self, query, *multiparams, **params):

        result = 0

        if self._readonly:
            raise MySQLReadOnlyError()

        if not isinstance(query, Update):
            raise TypeError(r'Not sqlalchemy.sql.dml.Update object')

        async with self._lock:

            proxy = await self._execute(query, *multiparams, **params)

            if proxy is not None:

                result = proxy.rowcount

                if not proxy.closed:
                    await proxy.close()

            if not self._trx:
                await self._close_conn()

        return result

    async def delete(self, query, *multiparams, **params):

        result = 0

        if self._readonly:
            raise MySQLReadOnlyError()

        if not isinstance(query, Delete):
            raise TypeError(r'Not sqlalchemy.sql.dml.Delete object')

        async with self._lock:

            proxy = await self._execute(query, *multiparams, **params)

            if proxy is not None:

                result = proxy.rowcount

                if not proxy.closed:
                    await proxy.close()

            if not self._trx:
                await self._close_conn()

        return result


class DBClient(_ClientBase, AsyncContextManager):
    """MySQL客户端对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    async def _context_release(self):

        await self._close_conn(self._lock.locked())

    async def release(self):

        async with self._lock:
            await self._close_conn()

    async def _execute(self, query, *multiparams, **params):

        global MYSQL_ERROR_RETRY_COUNT

        result = None

        async for times in AsyncCirculator(max_times=MYSQL_ERROR_RETRY_COUNT):

            try:

                conn = await self._get_conn()

                result = await conn.execute(query, *multiparams, **params)

            except (Warning, DataError, IntegrityError, ProgrammingError) as err:

                await self._close_conn(True)

                raise err

            except Exception as err:

                await self._close_conn(True)

                if times < MYSQL_ERROR_RETRY_COUNT:
                    Utils.log.error(err)
                else:
                    raise err

            else:

                break

        return result


class DBTransaction(_ClientBase, AsyncContextManager):
    """MySQL客户端事务对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    async def _get_conn(self):

        if self._trx is None:
            self._conn = await super()._get_conn()
            self._trx = await self._conn.begin()

        return self._conn

    async def _close_conn(self, discard=False):

        if self._trx is not None:

            if self._trx.is_active:
                self._trx.close()

            self._trx = None

        await super()._close_conn(discard)

        self._pool = None

    async def _context_release(self):

        await self.rollback()

    async def release(self):

        await self.rollback()

    async def _execute(self, query, *multiparams, **params):

        result = None

        if self._readonly:
            raise MySQLReadOnlyError()

        try:

            conn = await self._get_conn()

            result = await conn.execute(query, *multiparams, **params)

        except Exception as err:

            await self._close_conn(True)

            raise err

        return result

    async def commit(self):

        async with self._lock:

            if self._trx:
                await self._trx.commit()

            await self._close_conn()

    async def rollback(self):

        async with self._lock:

            if self._trx:
                await self._trx.rollback()

            await self._close_conn()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from motor.motor_asyncio import AsyncIOMotorClient

from .base import Utils

from ..error import catch_error


MONGO_POLL_WATER_LEVEL_WARNING_LINE = 0x10


class MongoPool:
    """Mongo连接管理
    """

    def __init__(
            self, host, username=None, password=None,
            *, name=None, auth_source=r'admin', min_pool_size=8, max_pool_size=32, max_idle_time=3600, wait_queue_timeout=10,
            compressors=r'zlib', zlib_compression_level=6,
            **settings
    ):

        self._name = name if name is not None else Utils.uuid1()[:8]

        settings[r'host'] = host
        settings[r'authSource'] = auth_source
        settings[r'minPoolSize'] = min_pool_size
        settings[r'maxPoolSize'] = max_pool_size
        settings[r'maxIdleTimeMS'] = max_idle_time * 1000
        settings[r'waitQueueTimeoutMS'] = wait_queue_timeout * 1000
        settings[r'compressors'] = compressors
        settings[r'zlibCompressionLevel'] = zlib_compression_level

        if username and password:
            settings[r'username'] = username
            settings[r'password'] = password

        self._pool = AsyncIOMotorClient(**settings)

        for server in self._servers.values():
            server.pool.remove_stale_sockets()

        self._pool_options = self._pool.options.pool_options

        Utils.log.info(
            f"Mongo {host} ({self._name}) initialized: "
            f"{self._pool_options.min_pool_size}/{self._pool_options.max_pool_size}"
        )

    @property
    def _servers(self):

        return self._pool.delegate._topology._servers

    def _echo_pool_info(self):

        global MONGO_POLL_WATER_LEVEL_WARNING_LINE

        for address, server in self._servers.items():

            active_size = server.pool.active_sockets
            max_pool_size = server.pool.max_pool_size

            if (max_pool_size - active_size) < MONGO_POLL_WATER_LEVEL_WARNING_LINE:
                Utils.log.warning(
                    f'Mongo connection pool not enough ({self._name}){address}: '
                    f'{active_size}/{max_pool_size}'
                )
            else:
                Utils.log.debug(
                    f'Mongo connection pool info ({self._name}){address}: '
                    f'{active_size}/{max_pool_size}'
                )

    def reset(self):

        for address, server in self._servers.items():

            server.pool.reset()
            server.pool.remove_stale_sockets()

            Utils.log.info(
                f'Mongo connection pool reset {address}: {len(server.pool.sockets)}/{self._pool.max_pool_size}'
            )

    def close(self):

        if self._pool is not None:
            self._pool.close()
            self._pool = None

    def get_database(self, db_name):

        self._echo_pool_info()

        result = None

        with catch_error():
            result = self._pool[db_name]

        return result


class MongoDelegate:
    """Mongo功能组件
    """

    def __init__(self):

        self._mongo_pool = None

    def init_mongo(self, *args, **kwargs):

        self._mongo_pool = MongoPool(*args, **kwargs)

    @property
    def mongo_pool(self):

        return self._mongo_pool

    async def mongo_health(self):

        result = False

        with catch_error():
            result = bool(await self._mongo_pool._pool.server_info())

        return result

    def reset_mongo_pool(self):

        self._mongo_pool.reset()

    def close_mongo_pool(self):

        self._mongo_pool.close()

    def get_mongo_database(self, db_name):

        return self._mongo_pool.get_database(db_name)

    def get_mongo_collection(self, db_name, collection):

        return self.get_mongo_database(db_name)[collection]



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'


import signal
import socket
import grpc
import msgpack

from ...extend.asyncio.base import Utils


UNARY_UNARY_MODE = grpc.unary_unary_rpc_method_handler
UNARY_STREAM_MODE = grpc.unary_stream_rpc_method_handler
STREAM_UNARY_MODE = grpc.stream_unary_rpc_method_handler
STREAM_STREAM_MODE = grpc.stream_stream_rpc_method_handler


class Router:

    def __init__(self, service):

        self._service = service
        self._handlers = {}

    @property
    def service(self):
        return self._service

    @property
    def handlers(self):
        return list(self._handlers.values())

    def register(self, func):

        if func.__name__ in self._handlers:
            raise ValueError(f'{func.__name__} has exists')

        self._handlers[func.__name__] = func

        return func


class GRPCServer:

    def __init__(
            self, interceptors=None, options=None, maximum_concurrent_rpcs=None, compression=None,
            request_deserializer=msgpack.loads, response_serializer=msgpack.dumps
    ):

        self._server: grpc.aio.Server = grpc.aio.server(
            interceptors=interceptors,
            options=options,
            maximum_concurrent_rpcs=maximum_concurrent_rpcs,
            compression=compression
        )

        self._request_deserializer = request_deserializer
        self._response_serializer = response_serializer

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    def _exit(self, *_):

        Utils.call_soon(self.stop)

    @property
    def server(self) -> grpc.aio.Server:

        return self._server

    def register(self, service, handlers, *, mode=UNARY_UNARY_MODE):

        generic_handlers = []

        for handler in handlers:
            generic_handlers.append(
                grpc.method_handlers_generic_handler(
                    service,
                    {
                        handler.__name__: mode(
                            handler,
                            request_deserializer=self._request_deserializer,
                            response_serializer=self._response_serializer,
                        )
                    }
                )
            )

        self._server.add_generic_rpc_handlers(generic_handlers)

    def bind_router(self, router: Router, *, mode=UNARY_UNARY_MODE):

        self.register(router.service, router.handlers, mode=mode)

    async def start(self, address, family=socket.AF_INET, server_credentials=None):

        if family == socket.AF_INET:
            _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            _address = f'{address[0]}:{address[1]}'
        elif family == socket.AF_INET6:
            _socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            _address = f'[{address[0]}]:{address[1]}'
        elif family == socket.AF_UNIX:
            _socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            _address = f'unix:{address}'
        else:
            raise ValueError(r'family invalid')

        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, True)
        _socket.bind(address)

        if server_credentials is None:
            self._server.add_insecure_port(_address)
        else:
            self._server.add_secure_port(_address, server_credentials)

        await self._server.start()

        Utils.log.success(f'grpc server [pid:{Utils.getpid()}] startup complete: {address}')

    async def stop(self, grace=None):

        await self._server.stop(grace)

    async def wait(self, timeout=None):

        return await self._server.wait_for_termination(timeout)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import pytz
import typing
import asyncio
import logging
import functools

from abc import abstractmethod

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from ..trace import refresh_trace_id
from ..interface import TaskInterface

from .base import Utils


TIMEZONE = pytz.timezone(r'Asia/Shanghai')


logging.getLogger(r'apscheduler').setLevel(logging.ERROR)


class TaskAbstract(TaskInterface):
    """任务基类
    """

    def __init__(self, scheduler=None):

        global TIMEZONE

        self._scheduler = AsyncIOScheduler(
            job_defaults={
                r'coalesce': False,
                r'max_instances': 1,
                r'misfire_grace_time': 10
            },
            timezone=TIMEZONE
        ) if scheduler is None else scheduler

    @property
    def scheduler(self):

        return self._scheduler

    @staticmethod
    def _func_wrapper(func, *args, **kwargs):

        @functools.wraps(func)
        async def _wrapper():
            refresh_trace_id()
            return await Utils.awaitable_wrapper(func(*args, **kwargs))

        return _wrapper

    def is_running(self):

        return self._scheduler.running

    def start(self):

        return self._scheduler.start()

    def stop(self):

        return self._scheduler.shutdown()

    @abstractmethod
    def add_job(self):

        raise NotImplementedError()

    def remove_job(self, job_id):

        return self._scheduler.remove_job(job_id)

    def remove_all_jobs(self):

        return self._scheduler.remove_all_jobs()


class IntervalTask(TaskAbstract):
    """间隔任务类
    """

    @classmethod
    def create(cls, interval, func, *args, **kwargs):

        inst = cls()

        inst.add_job(interval, func, *args, **kwargs)

        return inst

    def add_job(self, interval, func, *args, **kwargs):

        return self._scheduler.add_job(
            self._func_wrapper(func, *args, **kwargs),
            r'interval', seconds=interval
        )


class CronTask(TaskAbstract):
    """定时任务类
    """

    @classmethod
    def create(cls, crontab, func, *args, **kwargs):

        inst = cls()

        inst.add_job(crontab, func, *args, **kwargs)

        return inst

    def add_job(self, crontab, func, *args, **kwargs):

        return self._scheduler.add_job(
            self._func_wrapper(func, *args, **kwargs),
            CronTrigger.from_crontab(crontab, TIMEZONE)
        )


class DCSCronTask(TaskInterface):

    def __init__(self, redis_client, name, crontab, func, *args, **kwargs):

        self._name = name

        self._task_lock = redis_client.allocate_lock(f'dcs_cron_task:{name}', timeout=60)
        self._task_func = func

        self._heartbeat = IntervalTask.create(30, self._do_heartbeat)
        self._cron_task = CronTask.create(crontab, self._do_job, *args, **kwargs)

    @property
    def name(self):

        return self._name

    async def _do_heartbeat(self):

        resp = await self._check_status()

        Utils.log.debug(f'dcs cron task heartbeat: {self._name} => {resp}')

    async def _check_status(self):

        if await self._task_lock.acquire():
            return True
        else:
            return False

    async def _do_job(self, *args, **kwargs):

        if await self._check_status():

            Utils.log.info(f'dcs cron task start: {self._name}')

            if Utils.is_coroutine_function(self._task_func):
                await self._task_func(*args, **kwargs)
            else:
                self._task_func(*args, **kwargs)

            Utils.log.info(f'dcs cron task finish: {self._name}')

        else:

            Utils.log.debug(f'dcs cron task idle: {self._name}')

    def start(self, min_delay=30, max_delay=60):

        delay = Utils.randint(min_delay, max_delay)

        def _func():
            self._heartbeat.start()
            self._cron_task.start()
            Utils.log.info(f'dcs cron task init: {self._name}, delay: {delay}')

        if delay == 0:
            _func()
        else:
            Utils.call_later(delay, _func)

    def stop(self):

        self._heartbeat.stop()
        self._cron_task.stop()

    def is_running(self):

        return self._cron_task.is_running()


class MultiTasks:
    """多任务并发管理器

    提供协程的多任务并发的解决方案

    tasks = MultiTasks()
    tasks.append(func1())
    tasks.append(func2())
    ...
    tasks.append(funcN())
    await tasks

    多任务中禁止使用上下文资源共享的对象(如mysql和redis等)
    同时需要注意类似这种不能同时为多个协程提供服务的对象会造成不可预期的问题

    """

    def __init__(self):

        self._coroutines: typing.List[typing.Coroutine] = []

    def __await__(self) -> typing.List[typing.Any]:

        result = None

        if len(self._coroutines) > 0:

            result = yield from asyncio.gather(*self._coroutines).__await__()

            self._coroutines.clear()

        return result

    def __len__(self) -> int:

        return self._coroutines.__len__()

    def append(self, coroutine: typing.Coroutine):

        return self._coroutines.append(coroutine)

    def extend(self, coroutines: typing.List[typing.Coroutine]):

        return self._coroutines.extend(coroutines)

    def clear(self):

        for coroutine in self._coroutines:
            coroutine.close()

        self._coroutines.clear()


class SliceTasks(MultiTasks):
    """多任务分片并发管理器

    继承自MultiTasks类，通过参数tasks_num控制并发分片任务数

    """

    def __init__(self, tasks_num: int):

        super().__init__()

        self._tasks_num = max(1, tasks_num)

    def __await__(self) -> typing.List[typing.Any]:

        result = []

        if len(self._coroutines) > 0:

            for _idx in range(0, len(self._coroutines), self._tasks_num):

                tasks = self._coroutines[_idx: _idx + self._tasks_num]

                result.extend(
                    (yield from asyncio.gather(*tasks).__await__())
                )

            self._coroutines.clear()

        return result


class QueueTasks(MultiTasks):
    """多任务队列管理器

    继承自MultiTasks类，通过参数tasks_num控制队列长度

    """

    def __init__(self, tasks_num: int):

        super().__init__()

        self._tasks_num: int = max(1, tasks_num)

        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(self._tasks_num)

    def __await__(self) -> typing.List[typing.Any]:

        result = None

        if len(self._coroutines) > 0:

            tasks = [self._do_task(_task) for _task in self._coroutines]

            result = yield from asyncio.gather(*tasks).__await__()

            self._coroutines.clear()

        return result

    async def _do_task(self, coroutine: typing.Coroutine):

        async with self._semaphore:
            return await coroutine


class RateLimiter:
    """流量控制器，用于对计算资源的保护
    添加任务append函数如果成功会返回Future对象，可以通过await该对象等待执行结果
    进入队列的任务，如果触发限流行为会通过在Future上引发CancelledError传递出来
    """

    def __init__(self, running_limit: int, waiting_limit: int = 0, timeout: float = 0):

        self._timeout: float = timeout

        self._task_queue: asyncio.Queue[
            typing.Tuple[typing.Callable, typing.Tuple, typing.Dict, float]
        ] = asyncio.Queue(waiting_limit)

        self._consume_tasks: typing.List[asyncio.Task] = [
            asyncio.create_task(self._do_consume_task()) for _ in range(running_limit)
        ]

    def size(self) -> int:

        return self._task_queue.qsize()

    async def close(self):

        await self._task_queue.join()

        for task in self._consume_tasks:
            task.cancel()

        self._consume_tasks.clear()

    async def wait(self):

        return await self._task_queue.join()

    async def append(self, func: typing.Callable, *args, **kwargs) -> bool:

        result = False

        with Utils.suppress(asyncio.TimeoutError):

            if self._timeout > 0:

                await asyncio.wait_for(
                    self._task_queue.put(
                        (func, args, kwargs, Utils.loop_time())
                    ),
                    self._timeout
                )

            else:

                await self._task_queue.put(
                    (func, args, kwargs, Utils.loop_time())
                )

            result = True

        return result

    async def _do_consume_task(self):

        while True:

            try:

                func, args, kwargs, build_time = await self._task_queue.get()

                if (self._timeout > 0) and ((Utils.loop_time() - build_time) > self._timeout):
                    Utils.log.warning(f'rate limit timeout: {func} build_time:{build_time}')
                    continue

                await Utils.awaitable_wrapper(func(*args, **kwargs))

            except asyncio.CancelledError as err:

                raise err

            except Exception as err:

                Utils.log.error(str(err))

            finally:

                self._task_queue.task_done()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import pydantic
import fastapi

from inspect import Signature, Parameter


class ModelMetaclass(pydantic.main.ModelMetaclass):

    def __init__(cls, _what, _bases=None, _dict=None):

        super().__init__(_what, _bases, _dict)

        if cls.__fields__:

            _dict.setdefault(r'__annotations__', {})

            # 原始参数类型清单
            annotations = {}

            for _cls in _bases:
                annotations.update(_cls.__dict__.get(r'__annotations__', {}))
            else:
                annotations.update(_dict[r'__annotations__'])

            _dict[r'__annotations__'].clear()
            _dict[r'__annotations__'].update(annotations)

            func_params = []

            for field in cls.__fields__.values():

                func_params.append(
                    Parameter(
                        field.name, Parameter.POSITIONAL_OR_KEYWORD,
                        default=field.field_info,
                        annotation=annotations.get(field.name, Parameter.empty)
                    )
                )

            # 更新类方法
            cls.params = lambda **kwargs: kwargs
            cls.params.__signature__ = Signature(func_params)


class BaseModel(pydantic.main.BaseModel, metaclass=ModelMetaclass):
    pass


class Depends(fastapi.params.Depends):

    def __init__(self, dependency=None, *, use_cache=True):

        if hasattr(dependency, r'params') and callable(dependency.params):
            super().__init__(dependency=dependency.params, use_cache=use_cache)
        else:
            super().__init__(dependency=dependency, use_cache=use_cache)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import uvicorn

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../../'))

from fastapi.middleware.cors import CORSMiddleware
from fastapi_health import health

from hagworm.frame.fastapi.base import DEFAULT_HEADERS, create_fastapi
from hagworm.extend.base import Utils
from hagworm.extend.logging import LogFileRotator

from setting import ConfigDynamic
from service import DataSource
from controller import router


app = create_fastapi(
    log_level=ConfigDynamic.LogLevel,
    log_file_path=ConfigDynamic.LogFilePath,
    log_file_rotation=LogFileRotator.make(ConfigDynamic.LogFileSplitSize, ConfigDynamic.LogFileSplitTime),
    log_file_retention=ConfigDynamic.LogFileBackups,
    debug=ConfigDynamic.Debug,
    routes=router.routes,
    # on_startup=[DataSource.initialize],
    # on_shutdown=[DataSource.release],
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=ConfigDynamic.AllowOrigins,
    allow_methods=ConfigDynamic.AllowMethods,
    allow_headers=ConfigDynamic.AllowHeaders,
    allow_credentials=ConfigDynamic.AllowCredentials,
)

# app.add_api_route(r'/health', health([DataSource.check_health]))


if __name__ == r'__main__':

    Utils.log.warning(r'THE PRODUCTION ENVIRONMENT IS STARTED USING GUNICORN')

    uvicorn.run(app, port=ConfigDynamic.Port, log_config=None, headers=DEFAULT_HEADERS)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from typing import Any

from pydantic import errors
from pydantic.fields import ModelField

from ...extend import validator
from ...extend.base import Utils
from ...extend.struct import Enum


class _EnumKeyTypeBase(str):

    __subclasses__ = {}

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> int:

        if field.required is False and val == field.default:
            return val

        if not cls.Enum.has_key(val):
            raise errors.EnumMemberError(enum_values=cls.Enum)

        return getattr(cls.Enum, val).value


def EnumKeyType(enum):

    if enum not in _EnumKeyTypeBase.__subclasses__:

        _EnumKeyTypeBase.__subclasses__[enum] = type(
            f'EnumKeyType_{Utils.get_class_name(enum)}',
            (_EnumKeyTypeBase,),
            {r'Enum': enum}
        )

    return _EnumKeyTypeBase.__subclasses__.get(enum)


class _EnumValueTypeBase:

    __subclasses__ = {}
    __value_type__ = None

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> int:

        if field.required is False and val == field.default:
            return val

        if not isinstance(val, cls.__value_type__):
            val = cls.__value_type__(val)

        if not cls.Enum.has_value(val):
            raise errors.EnumMemberError(enum_values=cls.Enum)

        return val


def EnumValueType(enum: Enum):

    if enum not in _EnumValueTypeBase.__subclasses__:

        value_type = type(list(enum.values())[0])

        cls = _EnumValueTypeBase.__subclasses__[enum] = type(
            f'EnumValueType_{Utils.get_class_name(enum)}',
            (_EnumValueTypeBase, value_type),
            {r'Enum': enum}
        )

        cls.__value_type__ = value_type

    return _EnumValueTypeBase.__subclasses__.get(enum)


class IDCardType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not Utils.identity_card(val):
            raise ValueError(r'value is not a valid identity card')

        return val


class BankCardType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not Utils.luhn_valid(val):
            raise ValueError(r'value is not a valid bank card')

        return val


class UUIDType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.uuid(val):
            raise errors.UUIDError()

        return val


class DateType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> Any:

        if field.required is False and val == field.default:
            return val

        try:
            val = Utils.date_parse(val)
        except Exception:
            raise errors.DateError()

        return val


class JsonType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> Any:

        if field.required is False and val == field.default:
            return val

        try:
            val = Utils.json_decode(val)
        except Exception:
            raise errors.JsonError()

        return val


class IntListType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> Any:

        if field.required is False and val == field.default:
            return val

        try:
            val = Utils.split_int(val)
        except Exception:
            raise errors.ListError()

        return val


class FloatListType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> Any:

        if field.required is False and val == field.default:
            return val

        try:
            val = Utils.split_float(val)
        except Exception:
            raise errors.ListError()

        return val


class ASCVisibleType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.asc_visible(val):
            raise errors.StrRegexError(validator.REGEX_ASC_VISIBLE.pattern)

        return val


class EmailType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.email(val):
            raise errors.EmailError()

        return val


class DomainType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.domain(val):
            raise errors.UrlHostError()

        return val


class URLType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.url(val):
            raise errors.UrlSchemeError()

        return val


class MacAddrType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.mac_addr(val):
            raise ValueError(r'value is not a valid mac address')

        return val


class IPvAnyType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.ipv4(val) or not validator.ipv6(val):
            raise errors.IPvAnyAddressError()

        return val


class IPv4Type(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.ipv4(val):
            raise errors.IPv4AddressError()

        return val


class IPv4CidrType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.ipv4_cidr(val):
            raise errors.IPv4AddressError()

        return val


class IPv6Type(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.ipv6(val):
            raise errors.IPv6AddressError()

        return val


class IPv6CidrType(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Any, field: ModelField) -> str:

        if field.required is False and val == field.default:
            return val

        if not validator.ipv6_cidr(val):
            raise errors.IPv6AddressError()

        return val



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os

from hagworm.extend.base import Utils
from hagworm.extend.config import Configure, Field, HostType, StrListType


class _Static(Configure):

    MySqlMasterServer: HostType = Field(r'MySql')
    MySqlSlaveServer: HostType = Field(r'MySql')
    MySqlName: str = Field(r'MySql')
    MySqlUser: str = Field(r'MySql')
    MySqlPasswd: str = Field(r'MySql')
    MySqlMasterMinConn: int = Field(r'MySql')
    MySqlMasterMaxConn: int = Field(r'MySql')
    MySqlSlaveMinConn: int = Field(r'MySql')
    MySqlSlaveMaxConn: int = Field(r'MySql')

    MongoHost: StrListType = Field(r'Mongo')
    MongoAuth: str = Field(r'Mongo')
    MongoUser: str = Field(r'Mongo')
    MongoPasswd: str = Field(r'Mongo')
    MongoMinConn: int = Field(r'Mongo')
    MongoMaxConn: int = Field(r'Mongo')

    RedisHost: HostType = Field(r'Redis')
    RedisPasswd: str = Field(r'Redis')
    RedisMinConn: int = Field(r'Redis')
    RedisMaxConn: int = Field(r'Redis')
    RedisExpire: int = Field(r'Redis')
    RedisKeyPrefix: str = Field(r'Redis')


class _Dynamic(Configure):

    Port: int = Field(r'Base')
    Debug: bool = Field(r'Base')
    GZip: bool = Field(r'Base')
    Secret: str = Field(r'Base')
    ProcessNum: int = Field(r'Base')
    ServerName: str = Field(r'Base')

    AllowOrigins: StrListType = Field(r'Cros')
    AllowMethods: StrListType = Field(r'Cros')
    AllowHeaders: StrListType = Field(r'Cros')
    AllowCredentials: bool = Field(r'Cros')

    LogLevel: str = Field(r'Log')
    LogFilePath: str = Field(r'Log')
    LogFileSplitSize: int = Field(r'Log')
    LogFileSplitTime: str = Field(r'Log')
    LogFileBackups: int = Field(r'Log')

    ThreadPoolMaxWorkers: int = Field(r'ThreadPool')


ConfigStatic = _Static()
ConfigDynamic = _Dynamic()

cluster = Utils.getenv(r'CLUSTER', r'').strip().lower()

current_path = os.path.split(os.path.abspath(__file__))[0]

if cluster:
    ConfigStatic.read(os.path.join(current_path, f'./config/static.{cluster}.conf'))
    ConfigDynamic.read(os.path.join(current_path, f'./config/dynamic.{cluster}.conf'))
else:
    ConfigStatic.read(os.path.join(current_path, r'./config/static.conf'))
    ConfigDynamic.read(os.path.join(current_path, r'./config/dynamic.conf'))



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import binascii
import base64
import jwt
import textwrap

from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend


class RsaUtil:
    """Rsa加解密相关工具类
    """

    @classmethod
    def gen_rsa_key(cls, rsa_key, private=False):

        if private:
            start_line = r'-----BEGIN RSA PRIVATE KEY-----'
            end_line = r'-----END RSA PRIVATE KEY-----'
        else:
            start_line = r'-----BEGIN PUBLIC KEY-----'
            end_line = r'-----END PUBLIC KEY-----'

        rsa_key = textwrap.fill(rsa_key, 64)

        return '\n'.join([start_line, rsa_key, end_line])

    @classmethod
    def rsa_sign(cls, rsa_key, sign_data):

        algorithm = jwt.algorithms.RSAAlgorithm(hashes.SHA1)

        key = algorithm.prepare_key(cls.gen_rsa_key(rsa_key, True))

        signature = algorithm.sign(sign_data.encode(r'utf-8'), key)

        return base64.b64encode(signature).decode()

    @classmethod
    def rsa_verity(cls, pubic_key, verity_data, verity_sign):

        algorithm = jwt.algorithms.RSAAlgorithm(hashes.SHA1)

        public_key = load_pem_public_key(cls.gen_rsa_key(pubic_key).encode(r'utf-8'), backend=default_backend())

        result = algorithm.verify(verity_data.encode(), public_key, binascii.a2b_base64(verity_sign))

        return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import aiosmtplib

from email.message import EmailMessage

from .base import Utils


class EmailBody:

    def __init__(self, sender, recipients, message):

        self._sender = sender
        self._recipients = recipients
        self._message = message

    @property
    def sender(self):
        return self._sender

    @property
    def recipients(self):
        return self._recipients

    @property
    def message(self):
        return self._message

    @staticmethod
    def create_message(sender, recipients, subject, content, content_type=r'text/html'):

        message = EmailMessage()

        if sender is not None:
            message[r'From'] = sender

        if recipients is not None:
            message[r'To'] = recipients

        message[r'Subject'] = subject
        message.set_content(content)
        message.set_type(content_type)

        return message


class SMTPClient:

    def __init__(self, username, password, hostname, port, retry_count=5, **kwargs):

        self._username = username
        self._password = password

        self._hostname = hostname
        self._port = port

        self._retry_count = retry_count

        self._smtp_settings = kwargs

    def format_address(self, nickname, mailbox):

        return f'{nickname}<{mailbox}>'

    def format_addresses(self, items):

        return r';'.join(self.format_address(*item) for item in items)

    # 发送多封邮件
    async def send_messages(self, email_body_list):

        resp = None

        try:

            async with aiosmtplib.SMTP(hostname=self._hostname, port=self._port, **self._smtp_settings) as client:

                await client.login(self._username, self._password)

                for email_body in email_body_list:
                    resp = await client.send_message(email_body.message, email_body.sender, email_body.recipients)
                    Utils.log.info(f'{email_body.recipients} => {resp}')

        except aiosmtplib.SMTPException as err:

            Utils.log.error(err)

        return resp

    # 发送单封邮件
    async def send_message(self, sender, recipients, message):

        return await self.send_messages(
            [EmailBody(sender, recipients, message)]
        )

    # 简单的发送
    async def send(self, sender, recipients, subject, content):

        return await self.send_message(
            sender, recipients,
            EmailBody.create_message(sender, recipients, subject, content)
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import yarl
import asyncio
import aio_pika

from ...extend.asyncio.pool import ObjectInterface, ObjectPool


class RabbitMQConsumer(ObjectInterface):
    """RabbitMQ消费者
    """

    @staticmethod
    async def create_connection(
            url: yarl.URL, timeout: aio_pika.abc.TimeoutType = None, **kwargs
    ) -> aio_pika.RobustConnection:

        connection: aio_pika.RobustConnection = aio_pika.RobustConnection(url, **kwargs)

        await connection.connect(timeout)

        return connection

    def __init__(self, connection: aio_pika.abc.AbstractRobustConnection, queue_name: str):

        self._connection: aio_pika.abc.AbstractRobustConnection = connection
        self._channel: typing.Union[aio_pika.abc.AbstractRobustChannel, None] = None

        self._queue: typing.Union[aio_pika.abc.AbstractRobustQueue, None] = None
        self._queue_name: str = queue_name

    @property
    def queue_name(self) -> str:

        return self._queue_name

    async def open(
            self, *,
            consume_func=None, consume_no_ack=False, channel_qos_config=None, queue_config=None
    ):

        if channel_qos_config is None:
            channel_qos_config = {r'prefetch_count': 1}

        if queue_config is None:
            queue_config = {}

        self._channel = await self._connection.channel()

        await self._channel.set_qos(**channel_qos_config)

        self._queue = await self._channel.declare_queue(self._queue_name, **queue_config)

        if consume_func is not None:
            await self._queue.consume(consume_func, no_ack=consume_no_ack)

    async def close(self, with_connection=False):

        await self._channel.close()

        if with_connection is True:
            await self._connection.close()

    async def get(self, *, no_ack=False, timeout=1) -> aio_pika.abc.AbstractIncomingMessage:

        return await self._queue.get(no_ack=no_ack, timeout=timeout)


class RabbitMQConsumerForExchange(RabbitMQConsumer):
    """RabbitMQ注册到交换机的消费者
    """

    def __init__(self, connection: aio_pika.abc.AbstractRobustConnection, queue_name: str, exchange_name: str):
        
        super().__init__(connection, queue_name)

        self._exchange: typing.Union[aio_pika.abc.AbstractExchange, None] = None
        self._exchange_name: str = exchange_name

    @property
    def exchange_name(self) -> str:

        return self._exchange_name

    async def open(
            self, *,
            consume_func=None, consume_no_ack=False, channel_qos_config=None, queue_config=None, routing_key=None
    ):

        await super().open(
            consume_func=consume_func, consume_no_ack=consume_no_ack,
            channel_qos_config=channel_qos_config, queue_config=queue_config
        )

        self._exchange = await self._channel.get_exchange(self._exchange_name)

        await self._queue.bind(self._exchange, routing_key)

##################################################

class RabbitMQConsumerPool(ObjectPool):
    """RabbitMQ消费者连接池
    """

    def __init__(
            self, pool_size: int, url: yarl.URL, queue_name: str,
            timeout: aio_pika.abc.TimeoutType = None, **kwargs
    ):

        super().__init__(pool_size)

        self._queue_name: str = queue_name

        self._connection: aio_pika.RobustConnection = aio_pika.RobustConnection(url, **kwargs)

        asyncio.create_task(self._connection.connect(timeout))

    @property
    def queue_name(self) -> str:

        return self._queue_name

    def _create(self) -> RabbitMQConsumer:

        return RabbitMQConsumer(self._connection, self._queue_name)

    async def open(
            self, *,
            consume_func=None, consume_no_ack=False, channel_qos_config=None, queue_config=None,
    ):

        await self._connection.ready()

        await ObjectPool.open(
            self,
            consume_func=consume_func, consume_no_ack=consume_no_ack,
            channel_qos_config=channel_qos_config, queue_config=queue_config,
        )

    async def close(self):

        await super().close()

        await self._connection.close()

    async def get(self, *, no_ack=False, timeout=1) -> aio_pika.abc.AbstractIncomingMessage:

        async with self.get() as consumer:
            return await consumer.get(no_ack=no_ack, timeout=timeout)


class RabbitMQConsumerForExchangePool(RabbitMQConsumerPool):
    """RabbitMQ交换机消费者连接池
    """

    def __init__(
            self, pool_size: int, url: yarl.URL, queue_name: str, exchange_name: str,
            timeout: aio_pika.abc.TimeoutType = None, **kwargs
    ):

        super().__init__(pool_size, url, queue_name, timeout, **kwargs)

        self._exchange_name: str = exchange_name

    def _create(self) -> RabbitMQConsumerForExchange:

        return RabbitMQConsumerForExchange(self._connection, self._queue_name, self._exchange_name)

    async def open(
            self, *,
            consume_func=None, consume_no_ack=False, channel_qos_config=None, queue_config=None,
            routing_key=None
    ):

        await self._connection.ready()

        await ObjectPool.open(
            self,
            consume_func=consume_func, consume_no_ack=consume_no_ack,
            channel_qos_config=channel_qos_config, queue_config=queue_config, routing_key=routing_key
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from .base import Utils, FuncWrapper


class EventDispatcher:
    """事件总线
    """

    def __init__(self):

        self._observers = {}

    def _gen_observer(self):

        return FuncWrapper()

    def dispatch(self, _type, *args, **kwargs):

        if _type in self._observers:
            self._observers[_type](*args, **kwargs)

    def add_listener(self, _type, _callable):

        Utils.log.debug(f'add event listener => type({_type}) function({id(_callable)})')

        result = False

        if _type in self._observers:
            result = self._observers[_type].add(_callable)
        else:
            observer = self._observers[_type] = self._gen_observer()
            result = observer.add(_callable)

        return result

    def remove_listener(self, _type, _callable):

        Utils.log.debug(f'remove event listener => type({_type}) function({id(_callable)})')

        result = False

        if _type in self._observers:

            observer = self._observers[_type]

            result = observer.remove(_callable)

            if not observer.is_valid:
                del self._observers[_type]

        return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from coredis import lock, StrictRedis, StrictRedisCluster

from .base import Utils

from ..error import catch_error, catch_warning
from ..interface import AsyncContextManager
from .ntp import AsyncNTPClient
from .transaction import Transaction


REDIS_ERROR_RETRY_COUNT = 0x10
REDIS_POOL_WATER_LEVEL_WARNING_LINE = 0x10


class LuaLock(AsyncContextManager, lock.LuaLock):

    def __init__(
            self, redis, name, timeout=None, sleep=0.1,
            blocking=False, blocking_timeout=None, thread_local=True
    ):
        super().__init__(
            redis, name, timeout=timeout, sleep=sleep,
            blocking=blocking, blocking_timeout=blocking_timeout, thread_local=thread_local
        )

    async def _context_release(self):
        with catch_warning():
            await self.release()

    async def acquire(self, blocking=None, blocking_timeout=None):
        if self.local.get() is None:
            return await super().acquire(blocking, blocking_timeout)
        else:
            return await super().extend(self.timeout)

    async def release(self):
        if self.local.get() is not None:
            await super().release()


class ClusterLock(AsyncContextManager, lock.ClusterLock):

    def __init__(
            self, redis, name, timeout=None, sleep=0.1,
            blocking=False, blocking_timeout=None, thread_local=True
    ):
        super().__init__(
            redis, name, timeout=timeout, sleep=sleep,
            blocking=blocking, blocking_timeout=blocking_timeout, thread_local=thread_local
        )

    async def _context_release(self):
        with catch_warning():
            await self.release()

    async def acquire(self, blocking=None, blocking_timeout=None):
        if self.local.get() is None:
            return await super().acquire(blocking, blocking_timeout)
        else:
            return await super().extend(self.timeout)

    async def release(self):
        if self.local.get() is not None:
            await super().release()


class _PoolMixin(AsyncContextManager):

    def __init__(self, min_connections=0):

        self._name = Utils.uuid1()[:8]
        self._key_prefix = None
        self._min_connections = min_connections if min_connections else 0

    async def _context_release(self):

        pass

    def set_key_prefix(self, value):

        self._key_prefix = value

    def get_safe_key(self, key, *args, **kwargs):

        if self._key_prefix:
            _key = f'{self._key_prefix}:{key}'
        else:
            _key = key

        if args or kwargs:
            _key = f'{_key}:{Utils.params_sign(*args, **kwargs)}'

        return _key

    def reset(self):

        self.connection_pool.disconnect()

        self._init_connection()

        Utils.log.info(
            f"Redis connection pool reset ({self._name}): "
            f"{len(self.connection_pool._available_connections)}/{self.connection_pool.max_connections}"
        )

    def close(self):

        self.connection_pool.disconnect()
        self.connection_pool.reset()

    async def get_obj(self, name):

        result = await super().get(name)

        return Utils.pickle_loads(result) if result else result

    async def getset_obj(self, name, value):

        value = Utils.pickle_dumps(value)

        result = await super().getset(name, value)

        return Utils.pickle_loads(result) if result else result

    async def set_obj(self, name, value, ex=3600, px=None, nx=False, xx=False):

        value = Utils.pickle_dumps(value)

        return await super().set(name, value, ex=ex, px=px, nx=nx, xx=xx)


class StrictRedisPool(_PoolMixin, StrictRedis):
    """StrictRedis连接管理
    """

    def __init__(
            self,
            host, port=6379, db=0, password=None,
            min_connections=0, max_connections=32,
            max_idle_time=43200, idle_check_interval=1,
            **kwargs
    ):

        StrictRedis.__init__(
            self,
            host, port, db,
            password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            **kwargs
        )

        _PoolMixin.__init__(self, min(min_connections, max_connections))

    def _init_connection(self):

        connections = []

        with catch_error():
            for _ in range(self._min_connections):
                connections.append(
                    self.connection_pool.get_connection()
                )

        with catch_error():
            for connection in connections:
                self.connection_pool.release(connection)

    async def initialize(self):

        self._init_connection()

        config = self.connection_pool.connection_kwargs

        Utils.log.info(
            f"Redis Pool [{config[r'host']}:{config[r'port']}] ({self._name}) initialized: "
            f"{self.get_created_connections()}/{self.connection_pool.max_connections}"
        )

        return self

    def get_created_connections(self):

        return self.connection_pool._created_connections

    def allocate_lock(self, name, *, timeout=None, sleep=0.1, blocking_timeout=None, thread_local=True):

        return self.lock(
            name, timeout=timeout, sleep=sleep, blocking_timeout=blocking_timeout,
            lock_class=LuaLock, thread_local=thread_local
        )


class StrictRedisClusterPool(_PoolMixin, StrictRedisCluster):
    """StrictRedisCluster连接管理
    """

    def __init__(
            self,
            host=None, port=None, startup_nodes=None, password=None,
            min_connections=0, max_connections=32,
            max_idle_time=43200, idle_check_interval=1,
            **kwargs
    ):

        StrictRedisCluster.__init__(
            self,
            host, port, startup_nodes,
            password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            **kwargs
        )

        _PoolMixin.__init__(self, min(min_connections, max_connections))

    def _init_connection(self):

        connections = []

        with catch_error():
            for _ in range(self._min_connections):
                connections.append(
                    self.connection_pool.get_connection_by_key(Utils.uuid1())
                )

        with catch_error():
            for connection in connections:
                self.connection_pool.release(connection)

    async def initialize(self):

        await self.connection_pool.initialize()

        self._init_connection()

        nodes = self.connection_pool.nodes.nodes

        Utils.log.info(
            f"Redis Cluster Pool {list(nodes.keys())} ({self._name}) initialized: "
            f"{self.get_created_connections()}/{self.connection_pool.max_connections}"
        )

        return self

    def get_created_connections(self):

        return sum(self.connection_pool._created_connections_per_node.values())

    def allocate_lock(self, name, *, timeout=None, sleep=0.1, blocking_timeout=None, thread_local=True):

        return self.lock(
            name, timeout=timeout, sleep=sleep, blocking_timeout=blocking_timeout,
            lock_class=ClusterLock, thread_local=thread_local
        )


class RedisDelegate:
    """Redis功能组件
    """

    def __init__(self):

        self._redis_pool = None

    @property
    def redis_pool(self):

        return self._redis_pool

    async def init_redis_single(
            self,
            host, port=6379, db=0, password=None, *,
            min_connections=0, max_connections=32,
            max_idle_time=43200, idle_check_interval=1,
            **kwargs
    ):

        self._redis_pool = await StrictRedisPool(
            host, port, db, password=password,
            min_connections=min_connections, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            **kwargs
        ).initialize()

        return self._redis_pool

    async def init_redis_cluster(
            self,
            host=None, port=None, startup_nodes=None, password=None, *,
            min_connections=0, max_connections=32,
            max_idle_time=43200, idle_check_interval=1,
            **kwargs
    ):

        self._redis_pool = await StrictRedisClusterPool(
            host, port, startup_nodes, password=password,
            min_connections=min_connections, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            **kwargs
        ).initialize()

        return self._redis_pool

    def set_redis_key_prefix(self, value):

        self._redis_pool.set_key_prefix(value)

    async def redis_health(self):

        result = False

        with catch_error():
            result = bool(await self._redis_pool.info())

        return result

    def reset_redis_pool(self):

        self._redis_pool.reset()

    def close_redis_pool(self):

        self._redis_pool.close()

    def get_redis_client(self):

        return self._redis_pool


class ShareCache(AsyncContextManager):
    """共享缓存，使用with进行上下文管理

    基于分布式锁实现的一个缓存共享逻辑，保证在分布式环境下，同一时刻业务逻辑只执行一次，其运行结果会通过缓存被共享

    """

    def __init__(self, redis_client, share_key, lock_timeout=60, lock_blocking_timeout=60):

        self._redis_client = redis_client
        self._share_key = redis_client.get_safe_key(share_key)

        self._lock = self._redis_client.allocate_lock(
            redis_client.get_safe_key(f'share_cache:{share_key}'),
            timeout=lock_timeout, blocking_timeout=lock_blocking_timeout
        )

        self.result = None

    async def _context_release(self):

        await self.release()

    async def get(self):

        result = await self._redis_client.get_obj(self._share_key)

        if result is None:

            if await self._lock.acquire():
                result = await self._redis_client.get_obj(self._share_key)

        return result

    async def set(self, value, expire=None):

        return await self._redis_client.set_obj(self._share_key, value, expire)

    async def release(self):

        if self._lock:
            await self._lock.release()

        self._redis_client = self._lock = None


class PeriodCounter:

    MIN_EXPIRE = 60

    def __init__(self, redis_client, key_prefix, time_slice: int, *, ntp_client: AsyncNTPClient = None):

        self._redis_client = redis_client

        self._key_prefix = key_prefix
        self._time_slice = time_slice

        self._ntp_client = ntp_client

    def _get_key(self) -> str:

        timestamp = Utils.timestamp() if self._ntp_client is None else self._ntp_client.timestamp

        time_period = Utils.math.floor(timestamp / self._time_slice)

        return self._redis_client.get_safe_key(f'{self._key_prefix}:{time_period}')

    async def _execute(self, key: str, val: int) -> int:

        res = None

        async with await self._redis_client.pipeline(transaction=False) as pipeline:
            await pipeline.incrby(key, val)
            await pipeline.expire(key, max(self._time_slice, self.MIN_EXPIRE))
            res, _ = await pipeline.execute()

        return res

    async def incr(self, val: int = 1):

        return await self._execute(self._get_key(), val)

    async def incr_with_trx(self, val: int = 1) -> (int, Transaction):

        _key = self._get_key()

        res = await self._execute(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._execute, _key, -val)
        else:
            trx = None

        return res, trx

    async def decr(self, val: int = 1) -> int:

        return await self._execute(self._get_key(), -val)

    async def decr_with_trx(self, val: int = 1) -> (int, Transaction):

        _key = self._get_key()

        res = await self._execute(_key, -val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._execute, _key, val)
        else:
            trx = None

        return res, trx



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import time

from abc import abstractmethod

from texttable import Texttable

from ..extend.base import Utils
from ..extend.asyncio.socket import recv_json
from ..extend.asyncio.command import MainProcessAbstract, SubProcessAbstract


class TimerMS:

    __slots__ = [r'_timer']

    def __init__(self, timer=None):
        self._timer = time.time() if timer is None else timer

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.done()

    def done(self):
        self._timer = (time.time() - self._timer) * 1000

    @property
    def value(self):
        return self._timer


class Report:

    class _Report:

        def __init__(self):
            self.success = []
            self.failure = []

        @property
        def total(self):
            return len(self.success) + len(self.failure)

        @property
        def ratio(self):
            return len(self.success) / self.total if self.total > 0 else 0

        @property
        def min_success(self):
            return min(self.success)

        @property
        def max_success(self):
            return max(self.success)

        @property
        def avg_success(self):
            return sum(self.success) / len(self.success) if len(self.success) > 0 else 0

    def __init__(self):

        self._reports = {}

        self._timer = None

    def __enter__(self):

        self._timer = TimerMS()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self._timer.done()

        _time = max(round(self._timer.value / 1000), 1)
        _count = sum([item.total for item in self._reports.values()])

        table = Texttable()

        table.header(
            [
                r'Event Name',
                r'Success Total',
                r'Failure Total',
                r'Success Ratio',
                r'Success AveTime',
                r'Success MinTime',
                r'Success MaxTime',
            ]
        )

        for key, val in self._reports.items():
            table.add_row(
                [
                    key,
                    len(val.success),
                    len(val.failure),
                    r'{:.2%}'.format(val.ratio),
                    r'{:.3f}ms'.format(val.avg_success),
                    r'{:.3f}ms'.format(val.min_success),
                    r'{:.3f}ms'.format(val.max_success),
                ]
            )

        Utils.log.info(f'\nTotal: {_count}, Time: {_time}s, Qps: {round(_count / _time)}\n{table.draw()}\n')

    def _get_report(self, name: str) -> _Report:

        if name not in self._reports:
            self._reports[name] = self._Report()

        return self._reports[name]

    def add(self, name, result, resp_time):

        if result is True:
            self._get_report(name).success.append(resp_time)
        else:
            self._get_report(name).failure.append(resp_time)


class RunnerAbstract(SubProcessAbstract):

    @abstractmethod
    async def _run(self):
        raise NotImplementedError()

    async def success(self, name: str, resp_time: int):

        await self._socket_client.send_json(
            [name, True, resp_time]
        )

    async def failure(self, name: str, resp_time: int):

        await self._socket_client.send_json(
            [name, False, resp_time]
        )


class Launcher(MainProcessAbstract):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._report = Report()

    async def _client_connected_cb(self, reader, writer):

        while True:

            request = await recv_json(reader)

            if request:
                self._report.add(*request)
            else:
                break

    async def _run(self):

        with self._report:
            await super()._run()
